#![allow(unused, clippy::unnecessary_fallible_conversions)]
use std::path::Path;

use anyhow::Context as _;
use core::prelude::rust_2021::TryFrom;
use std::path::PathBuf;
use tracing_subscriber::prelude::*;

use anyhow::Result;
use tracing::{debug, error, field::Empty, info, instrument, warn};
use tracing::{level_filters::LevelFilter, Span};

fn main() {
    tracing_subscriber::registry()
        .with(tracing_subscriber::fmt::layer().with_filter(LevelFilter::DEBUG))
        .init();

    no_tracing::main();
    with_anyhow_tracing::main();
}

mod no_tracing {
    use super::*;

    pub fn main() {
        if let Err(e) = do_something("file.txt") {
            warn!("Couldn't do the thing: {e:#}");
            if let Err(e) = do_something_else("other_file.txt") {
                error!("Couldn't do the other thing: {e:#}");
                return;
            }
        }
        info!("All done");
    }

    fn do_something(path: &str) -> Result<()> {
        info!("Doing something to \"{path}\"");

        let path = PathBuf::from(path)
            .canonicalize()
            .with_context(|| format!("Bad path {path:?}"))?;

        let mut data = get_data().context("Couldn't get data")?;
        debug!("Got data: {data:?}");

        change_data(&mut data).with_context(|| format!("Failed to change data {data:?}"))?;

        write_to_a_file(&path, &data).with_context(|| format!("Couldn't save data {data:?}"))?;
        debug!("Did something to {path:?}");
        Ok(())
    }

    fn get_data() -> Option<Vec<u8>> {
        Some(vec![1, 2, 3])
    }
    fn change_data(data: &mut [u8]) -> Result<()> {
        for i in data.iter_mut() {
            *i += 1;
        }
        debug!("Modifying data");
        Ok(())
    }

    fn write_to_a_file(path: &Path, data: &[u8]) -> Result<()> {
        std::fs::write(path, data).with_context(|| format!("Failed to write to file: {path:?}"))?;
        debug!(target: "fs", "Wrote to {path:?}");
        Ok(())
    }

    fn do_something_else(path: &str) -> Result<()> {
        info!("Doing something else");

        let path = PathBuf::try_from(path).with_context(|| format!("Bad path {path:?}"))?;

        write_to_a_file(&path, b"couldn't do the thing")?;
        Ok(())
    }
}

mod with_anyhow_tracing {
    use tracing_subscriber::field::debug;

    use super::*;

    pub fn main() {
        if let Err(e) = do_something("file.txt") {
            e.span().in_scope(|| warn!("{e:#}"));
            if let Err(e) = do_something_else("other_file.txt") {
                e.span().in_scope(|| error!("{e:#}"));
                return;
            }
        }
        info!("All done");
    }

    #[instrument(level = "warn", fields(data = Empty, new_data = Empty))]
    fn do_something(path: &str) -> Result<()> {
        info!("Doing something");

        let path = PathBuf::from(path).canonicalize().context("Bad path")?;

        let mut data = get_data().context("Couldn't get data")?;
        Span::current().record("data", format!("{data:?}"));
        debug!("Got data");

        change_data(&mut data).context("Failed to change data")?;
        Span::current().record("new_data", format!("{data:?}"));
        debug!("Changed data");

        write_to_a_file(&path, &data)?;
        debug!("Did something");
        Ok(())
    }

    fn get_data() -> Option<Vec<u8>> {
        Some(vec![1, 2, 3])
    }
    fn change_data(data: &mut [u8]) -> Result<()> {
        for i in data.iter_mut() {
            *i += 1;
        }
        debug!("Modifying data");
        Ok(())
    }

    #[instrument(skip_all, target = "fs" fields(path = %path.display()))]
    fn write_to_a_file(path: &Path, data: &[u8]) -> Result<()> {
        std::fs::write(path, data).context("Failed to write to file")?;
        debug!(target: "fs", "Wrote file");
        Ok(())
    }

    #[instrument(level = "warn", skip_all)]
    fn do_something_else(path: &str) -> Result<()> {
        info!("Doing something else");
        let path = PathBuf::from(path).canonicalize().context("Bad path")?;
        write_to_a_file(&path, b"couldn't do the thing")?;
        Ok(())
    }
}
